from ..builders import OutputBuilder
from ..outputs.toml import TOMLOutput


class InvenioRecordMetadataModelsPoetryBuilder(OutputBuilder):
    TYPE = 'invenio_record_metadata_models_poetry'

    def finish(self):
        super().finish()

        output: TOMLOutput = self.builder.get_output(
            'toml',
            'pyproject.toml'
        )

        metadata_package = self.settings.python.record_metadata_class.rsplit('.', maxsplit=1)[0]

        output.set(
            "tool.poetry.plugins.'invenio_db.models'",
            self.settings.python.record_schema_metadata_poetry,
            metadata_package
        )



from collections import defaultdict, namedtuple

from libcst import CSTTransformer, ClassDef, FunctionDef, SimpleStatementLine, Import, ImportFrom, Assign, List, Expr, \
    IndentedBlock, Element, Pass


def merge(existing_cst, new_cst, top_cst=None, mergers=None, **kwargs):
    return existing_cst.visit(MergingTransformer(top_cst or new_cst, new_cst, mergers or general_mergers, **kwargs))


node_with_type = namedtuple('node_with_type', 'node, type')


class MergingTransformer(CSTTransformer):
    def __init__(self, top_cst, new_cst, mergers, finalizer=None):
        super().__init__()
        self.top_cst = top_cst
        self.new_cst = new_cst
        self.mergers = mergers
        self.finalizer = finalizer
        self.node_type_category = {
            Import: 'import',
            ImportFrom: 'import',
            ClassDef: 'classdef',
            Assign: 'assign'
        }

    def on_visit(self, node):  # do not process children
        return False

    def on_leave(self, original_node, updated_node):
        existing_list = self.extract_body(updated_node)
        new_list = self.extract_body(self.new_cst)

        ret = []
        existing_list = [node_with_type(e, self.get_node_type(e)) for e in existing_list]
        new_list = [node_with_type(e, self.get_node_type(e)) for e in new_list]

        last_type = None
        for existing in existing_list:
            if last_type is not None and last_type != existing.type:
                while new_list and new_list[0].type == last_type:
                    ret.append(new_list.pop().node)
            last_type = existing.type
            found = False
            for idx, new in enumerate(new_list):
                if new.type != existing.type:
                    break
                if type(self.real_node(new.node)) is type(self.real_node(existing.node)):
                    merger = self.mergers.get(type(self.real_node(existing.node)), IdentityMerger())
                    if merger.should_merge(self.top_cst, existing.node, new.node):
                        ret.append(merger.merge(self.top_cst, existing.node, new.node))
                        del new_list[idx]
                        found = True
                        break
            if not found:
                ret.append(existing.node)

        for node in new_list:
            ret.append(node.node)

        if self.finalizer:
            return self.finalizer(updated_node, ret)
        else:
            return updated_node.with_changes(
                body=ret
            )

    def real_node(self, node):
        if isinstance(node, SimpleStatementLine):
            return node.body[0]
        return node

    def get_node_type(self, node):
        t = type(self.real_node(node))
        return self.node_type_category.get(t, 'unknown')

    def extract_body(self, node):
        ret = defaultdict(list)
        if hasattr(node, 'body'):
            body = node.body
        elif isinstance(node, List):
            body = node.elements
        else:
            raise Exception(f'Do not know how to get body from {node}')
        if isinstance(body, IndentedBlock):
            body = body.body
        return body


class MergerBase:
    def merge(self, cst, existing_node, new_node):
        raise NotImplementedError()

    def should_merge(self, cst, existing_node, new_node):
        return False

    def real_node(self, node):
        if isinstance(node, SimpleStatementLine):
            return node.body[0]
        return node

    def real_with_changes(self, node, **kwargs):
        if isinstance(node, SimpleStatementLine):
            return node.with_changes(
                body=[
                    node.body[0].with_changes(**kwargs)
                ]
            )
        else:
            return node.with_changes(**kwargs)


class IdentityBaseMerger(MergerBase):
    def merge(self, cst, existing_node, new_node):
        return existing_node

    def should_merge(self, cst, existing_node, new_node):
        return cst.code_for_node(existing_node).strip() == cst.code_for_node(new_node).strip()


class IdentityMerger(IdentityBaseMerger):
    def merge(self, cst, existing_node, new_node):
        return existing_node

    def should_merge(self, cst, existing_node, new_node):
        print(
            f'Warning: using naive merger for {type(self.real_node(existing_node))}'
            f'>>>\n{cst.code_for_node(existing_node).strip()}\n<<<')
        return super().should_merge(cst, existing_node, new_node)


class ClassMerger(MergerBase):
    def merge(self, cst, existing_node, new_node):
        return merge(existing_node, new_node, cst, general_mergers, finalizer=self.finalize)

    def finalize(self, node, new_content):
        return node.with_changes(
            body=node.body.with_changes(body=new_content)
        )

    def should_merge(self, cst, existing_node, new_node):
        return existing_node.name.value == new_node.name.value


class AssignMerger(MergerBase):
    def merge(self, cst, existing_node, new_node):
        # TODO: merge dictionaries
        real_existing = self.real_node(existing_node)
        real_new = self.real_node(new_node)
        if isinstance(real_existing.value, List) or isinstance(real_new.value, List):
            return self.real_with_changes(existing_node, value=self.merge_lists(
                cst, real_existing.value, real_new.value
            ))
        return existing_node

    def merge_lists(self, cst, existing_list, new_list):
        return merge(existing_list, new_list, cst, list_mergers, finalizer=self.finalize_list)

    def finalize_list(self, node, new_content):
        return node.with_changes(elements=new_content)

    def node_to_name(self, node):
        return tuple(sorted([x.target.value for x in self.real_node(node).targets]))

    def should_merge(self, cst, existing_node, new_node):
        return self.node_to_name(existing_node) == self.node_to_name(new_node)


class ImportMerger(IdentityBaseMerger):
    pass


class ImportFromMerger(IdentityBaseMerger):
    pass


class ExprMerger(IdentityBaseMerger):
    pass


class PassMerger(IdentityBaseMerger):
    pass


class FunctionMerger(MergerBase):
    def merge(self, cst, existing_node, new_node):
        return existing_node

    def should_merge(self, cst, existing_node, new_node):
        return existing_node.name.value == new_node.name.value


class ElementMerger(MergerBase):
    def merge(self, cst, existing_node, new_node):
        return existing_node

    def should_merge(self, cst, existing_node, new_node):
        return cst.code_for_node(existing_node.value).strip() == cst.code_for_node(new_node.value).strip()


general_mergers = {
    ClassDef: ClassMerger(),
    Assign: AssignMerger(),
    Import: ImportMerger(),
    ImportFrom: ImportFromMerger(),
    Expr: ExprMerger(),
    FunctionDef: FunctionMerger(),
    Pass: PassMerger()
}

list_mergers = {
    Element: ElementMerger()
}



from collections import defaultdict

from oarepo_model_builder.builders import process
from oarepo_model_builder.stack import ModelBuilderStack
from .invenio_base import InvenioBaseClassPythonBuilder
from ..outputs.json_stack import JSONStack
from ..utils.deepmerge import deepmerge
from ..utils.jinja import package_name, base_name, with_defined_prefix
from ..utils.schema import is_schema_element, match_schema, Ref

OAREPO_MARSHMALLOW_PROPERTY = 'oarepo:marshmallow'


class MarshmallowNode:
    def __init__(self, schema_class_name, schema_class_bases, schema):
        self.schema_class_name = schema_class_name
        self.schema_class_bases = schema_class_bases
        self.schema = schema
        self.fields = {}
        self.prepared_schema = None

    def add(self, field_name, field_type):
        self.fields[field_name] = field_type

    def pop_prepared_schema(self):
        ret = self.prepared_schema
        self.prepared_schema = None
        return ret


class InvenioRecordSchemaBuilder(InvenioBaseClassPythonBuilder):
    TYPE = 'invenio_record_schema'
    class_config = 'record-schema-class'
    template = 'record-schema'

    def begin(self, schema, settings):
        super().begin(schema, settings)
        self.marshmallow_stack = []
        self.imports = defaultdict(set)  # import => aliases
        self.imports['marshmallow'].add('ma')
        self.imports['marshmallow.fields'].add('ma_fields')
        self.imports['marshmallow.validate'].add('ma_valid')
        self.imported_classes = {}

    def finish(self):
        # TODO: generate subschemas as well
        # TODO: generate arrays as well
        # TODO: handle required
        super().finish(
            imports=self.imports,
            imported_classes=self.imported_classes
        )

    @process('/model/**', condition=lambda current, stack: is_schema_element(stack))
    def enter_model_element(self, stack: ModelBuilderStack):
        schema_path = match_schema(stack)
        if isinstance(schema_path[-1], Ref):
            schema_element_type = schema_path[-1].element_type
        else:
            schema_element_type = None

        definition = None
        recurse = True
        if schema_element_type == 'properties':
            parent = stack[-2].data
            definition = parent.get(OAREPO_MARSHMALLOW_PROPERTY, {})

            # set nested if not already set
            if 'nested' not in definition:
                definition['nested'] = True

            generate_schema_class = False
            if 'class' in definition:
                schema_class = definition['class']
                if '.' not in schema_class:
                    schema_class = package_name(
                        self.settings.python.record_schema_class) + '.' + schema_class
                schema_class_base_classes = definition.get('base-classes', ['ma.Schema'])
                generate_schema_class = definition.get('generate', False)
            if generate_schema_class:
                self.marshmallow_stack.append(
                    MarshmallowNode(schema_class, schema_class_base_classes, stack.top.data))
                # add nested to the definition
            else:
                schema_element_type = 'nodef'
            recurse = generate_schema_class

        if recurse:
            # process children
            yield
        else:
            # skip children
            yield stack.SKIP

        data = stack.top.data

        if not self.marshmallow_stack:
            return

        marshmallow_stack_top = self.marshmallow_stack[-1]

        if schema_element_type == 'nodef':
            definition = self.get_marshmallow_definition(data, stack, definition=definition)
            self.marshmallow_stack[-1].prepared_schema = definition
        elif schema_element_type == 'items':
            definition = self.get_marshmallow_definition(data, stack)
            definition['field'] = f"ma.List({definition['field']})"
            self.marshmallow_stack[-1].prepared_schema = definition
        elif schema_element_type == 'property':
            prepared_schema = marshmallow_stack_top.pop_prepared_schema()
            if prepared_schema:
                deepmerge(data.setdefault(OAREPO_MARSHMALLOW_PROPERTY, {}), prepared_schema)
            definition = self.get_marshmallow_definition(data, stack)
            key = definition.get('field_name', stack.top.key)
            marshmallow_stack_top.add(key, definition['field'])
        elif schema_element_type == 'properties':
            node = self.marshmallow_stack.pop()
            self.process_template(
                self.class_to_path(node.schema_class_name),
                'object-schema',
                schema_class=node.schema_class_name,
                schema_bases=node.schema_class_bases,
                fields=node.fields,
                imports=self.imports,
                imported_classes=self.imported_classes,
                current_package_name=package_name(node.schema_class_name)
            )

    def get_marshmallow_definition(self, data, stack, definition=None):
        if not definition:
            definition = data.setdefault(OAREPO_MARSHMALLOW_PROPERTY, {})

        definition = self.call_components('invenio_before_set_marshmallow_definition', definition, stack=stack)

        data_type = data.get('type', 'object')
        generator = self.schema.settings.python.marshmallow.mapping.get(data_type, None)
        if isinstance(generator, dict):
            # a value, not a callable => merge it
            definition = deepmerge(generator, definition)
            data['marshmallow'] = definition
            generator = None

        # add imports if required
        for imp in definition.get('imports', []):
            self.imports[imp['import']].add(imp['alias'])

        self.imported_classes.update(definition.get('imported-classes', {}))

        if 'field' in definition:
            # do not modify the field
            return definition

        if 'class' in definition:
            class_name = definition['class']
            if '.' in class_name:
                if not with_defined_prefix(self.settings.python.always_defined_import_prefixes, class_name):
                    class_base_name = self.imported_classes[definition['class']] = base_name(class_name)
                else:
                    class_base_name = class_name
            else:
                class_base_name = class_name

            # generate instance of the class, filling the options and validators

            definition['field'] = create_field(class_base_name, options=(), validators=(),
                                               definition=definition)
            return definition

        # if no generator from settings, get the default one
        if not generator:
            generator = default_marshmallow_generators.get(data_type, None)
            if not generator:
                if data_type == 'object':
                    raise Exception(
                        f'Do not have marshmallow field generator for type "{data_type}" at path "{stack.path}". '
                        f'Either supply an existing schema class or instruct the compiler to generate one. '
                        f'See the documentation (docs/model.md, section oarepo:marshmallow) for details.')
                raise Exception(
                    f'Do not have marshmallow field generator for type "{data_type}" at path "{stack.path}". '
                    f'Define it either in invenio_record_schema.py or in your own config')

        # and generate the field
        definition['field'] = generator(data, definition, self.schema, self.imports)
        definition = self.call_components('invenio_after_set_marshmallow_definition', definition, stack=stack)
        return definition


def create_field(field_type, options=(), validators=(), definition=None):
    opts = [*options, *definition.get('options', [])]
    validators = [*validators, *definition.get('validators', [])]
    nested = definition.get('nested', False)
    if validators:
        opts.append(f'validate=[{",".join(validators)}]')
    kwargs = definition.get('field_args', '')
    if kwargs and opts:
        kwargs = ', ' + kwargs
    if not nested:
        ret = f'{field_type}({", ".join(opts)}{kwargs})'
    else:
        ret = f'{field_type}'
    if nested:
        if opts or kwargs:
            ret = f'ma_fields.Nested({ret}, {", ".join(opts)}{kwargs})'
        else:
            ret = f'ma_fields.Nested({ret})'
    return ret


def marshmallow_string_generator(data, definition, schema, imports):
    validators = []
    min_length = data.get('minLength', None)
    max_length = data.get('maxLength', None)
    if min_length is not None or max_length is not None:
        validators.append(f'ma_valid.Length(min={min_length}, max={max_length})')
    return create_field('ma_fields.String', [], validators, definition)


def marshmallow_integer_generator(data, definition, schema, imports):
    validators = []
    minimum = data.get('minimum', None)
    maximum = data.get('maximum', None)
    exclusive_minimum = data.get('exclusiveMinimum', None)
    exclusive_maximum = data.get('exclusiveMaximum', None)

    if minimum is not None or maximum is not None or exclusive_minimum is not None or exclusive_maximum is not None:
        validators.append(
            f'ma_valid.Range('
            f'min={minimum or exclusive_minimum or "None"}, max={maximum or exclusive_maximum or "None"}, '
            f'min_inclusive={exclusive_minimum is None}, max_inclusive={exclusive_maximum is None})')
    return create_field('ma_fields.Integer', [], validators, definition)


# TODO: rest of supported schema types

default_marshmallow_generators = {
    'string': marshmallow_string_generator,
    'integer': marshmallow_integer_generator
}



from .invenio_base import InvenioBaseClassPythonBuilder


class InvenioRecordDumperBuilder(InvenioBaseClassPythonBuilder):
    TYPE = 'invenio_record_dumper'
    class_config = 'record-dumper-class'
    template = 'record-dumper'



from .invenio_base import InvenioBaseClassPythonBuilder

class InvenioRecordPermissionsBuilder(InvenioBaseClassPythonBuilder):
    TYPE = 'invenio_record_permissions'
    class_config = 'record-permissions-class'
    template = 'record-permissions'



import datetime
import json
import logging
import os
import sys
from pathlib import Path

import click
import yaml

from oarepo_model_builder.entrypoints import load_entry_points_dict, create_builder_from_entrypoints, \
    load_included_models_from_entry_points, load_model
from oarepo_model_builder.schema import ModelSchema
from oarepo_model_builder.utils.hyphen_munch import HyphenMunch
from oarepo_model_builder.utils.verbose import log


@click.command()
@click.option('--output-directory',
              help='Output directory where the generated files will be placed. '
                   'Defaults to "."')
@click.option('--package',
              help='Package into which the model is generated. '
                   'If not passed, the name of the current directory, '
                   'converted into python package name, is used.')
@click.option('--set', 'sets',
              help='Overwrite option in the model file. Example '
                   '--set name=value',
              multiple=True)
@click.option('-v', 'verbosity',
              help='Increase the verbosity. This option can be used multiple times.',
              count=True)
@click.option('--config', 'configs',
              help='Load a config file and replace parts of the model with it. '
                   'The config file can be a json, yaml or a python file. '
                   'If it is a python file, it is evaluated with the current '
                   'model stored in the "oarepo_model" global variable and '
                   'after the evaluation all globals are set on the model.',
              multiple=True)
@click.option('--isort/--skip-isort', default=True, help='Call isort on generated sources')
@click.option('--black/--skip-black', default=True, help='Call black on generated sources')
@click.argument('model_filename')
def run(output_directory, package, sets, configs, model_filename, verbosity, isort, black):
    """
    Compiles an oarepo model file given in MODEL_FILENAME into an Invenio repository model.
    """

    # extend system's search path to add script's path in front (so that scripts called from the compiler are taken
    # from the correct virtual environ)
    os.environ['PATH'] = str(Path(sys.argv[0]).parent.absolute()) + os.pathsep + os.environ.get('PATH', '')

    if not output_directory:
        output_directory = os.getcwd()

    # set the logging level, it will be warning - 1 (that is, 29) if not verbose,
    # so that warnings only will be emitted. With each verbosity level
    # it will decrease
    logging.basicConfig(
        level=logging.INFO - verbosity,
        format=''
    )

    Path(output_directory).mkdir(parents=True, exist_ok=True)
    handler = logging.FileHandler(Path(output_directory) / 'installation.log', 'a')
    handler.setLevel(logging.INFO)
    logging.root.addHandler(handler)

    log.enter(0, '\n\n%s\n\nProcessing model %s into output directory %s',
              datetime.datetime.now(), model_filename, output_directory)

    schema = load_model(model_filename, package, configs, black, isort, sets)
    schema.schema['output-directory'] = output_directory
    builder = create_builder_from_entrypoints()
    builder.build(schema, output_directory)

    log.leave('Done')

    print(f"Log saved to {Path(output_directory) / 'installation.log'}")


if __name__ == '__main__':
    run()



from .invenio_base import InvenioBaseClassPythonBuilder


class InvenioConfigBuilder(InvenioBaseClassPythonBuilder):
    TYPE = 'invenio_config'
    class_config = 'config-dummy-class'
    template = 'config'



from ..builders import OutputBuilder
from ..outputs.toml import TOMLOutput


class InvenioExtPoetryBuilder(OutputBuilder):
    TYPE = 'invenio_ext_poetry'

    def finish(self):
        super().finish()

        output: TOMLOutput = self.builder.get_output(
            'toml',
            'pyproject.toml'
        )

        ext_class = self.settings.python.ext_class.rsplit('.', maxsplit=1)

        output.set("tool.poetry.plugins.'invenio_base.api_apps'",
                   self.settings.package,
                   f'{ext_class[0]}:{ext_class[-1]}'
                   )



import copy
from typing import List

from oarepo_model_builder.utils.deepmerge import deepmerge


class JSONStack:
    """Hierarchic json builder."""
    IGNORED_SUBTREE = object()
    IGNORED_NODE = object()

    def __init__(self):
        self.stack: List = [{}]

    def should_ignore(self, element):
        return element is self.IGNORED_SUBTREE or element is self.IGNORED_NODE

    def push(self, key, el):
        if key is None:
            assert isinstance(el, dict)
            assert self.empty
            self.stack[0] = deepmerge(copy.deepcopy(el), self.stack[0], [])
            return
        top = self.stack[-1]
        if top is self.IGNORED_SUBTREE:
            self.stack.append(self.IGNORED_SUBTREE)
        elif self.should_ignore(el):
            self.stack.append(el)
        else:
            if top is self.IGNORED_NODE:
                top = self.real_top

            el = copy.deepcopy(el)
            if isinstance(top, dict):
                if key not in top:
                    top[key] = el
                else:
                    top[key] = deepmerge(el, top[key])
            elif isinstance(top, list):
                if key < len(top):
                    top[key] = deepmerge(el, top[key])
                else:
                    assert key == len(top)
                    top.append(el)
            else:
                raise NotImplemented(f'Set for datatype {type(top)} is not implemented')
            self.stack.append(el)

    def pop(self):
        if not self.empty:
            self.stack.pop()

    @property
    def empty(self):
        return len(self.stack) == 1

    @property
    def value(self):
        return self.stack[0]

    @property
    def real_top(self):
        for t in reversed(self.stack):
            if t is not self.IGNORED_NODE:
                return t

    def merge(self, value):
        real_top = self.real_top
        real_top.update(deepmerge(value, real_top, []))



from .invenio_base import InvenioBaseClassPythonBuilder


class InvenioViewsBuilder(InvenioBaseClassPythonBuilder):
    TYPE = 'invenio_views'
    class_config = 'create_blueprint_from_app'
    template = 'views'



from oarepo_model_builder.stack import ModelBuilderStack


class SchemaPathValidator:
    valid = False

    def get(self, path_el):
        raise NotImplementedError()

    @property
    def real(self):
        return self


class Invalid(SchemaPathValidator):
    def get(self, path_el):
        return self


invalid = Invalid()


class PrimitiveValidator(SchemaPathValidator):
    valid = True

    def get(self, path_el):
        return invalid


primitive = PrimitiveValidator()


class DictValidator(SchemaPathValidator):
    valid = True

    def __init__(self, dict=None, primitives=None):
        self.dict = dict or {}
        if primitives:
            for p in primitives.split(','):
                p = p.strip()
                if not p:
                    continue
                self.dict[p] = primitive

    def get(self, path_el):
        if path_el in self.dict:
            return self.dict[path_el]
        return invalid


class AnyKeyDictValidator(SchemaPathValidator):
    valid = True

    def __init__(self, _next):
        self._next = _next

    def get(self, path_el):
        return self._next


class Ref(SchemaPathValidator):
    refs = {}
    valid = True

    def __init__(self, element_type, refstr):
        self.element_type = element_type
        self.refstr = refstr

    def get(self, path_el):
        return self.refs[self.refstr].get(path_el)

    @property
    def real(self):
        return self.refs[self.refstr]

    def __str__(self):
        return self.element_type

    def __repr__(self):
        return str(self)


Ref.refs['type'] = DictValidator(dict={
    "properties": Ref('properties', 'properties'),
    "patternProperties": Ref('patternProperties', 'properties'),
    "additionalProperties": Ref('additionalProperties', 'additionalProperties'),
    "propertyNames": Ref('propertyNames', 'propertyNames'),
    "items": Ref('items', 'type'),
    "prefixItems": Ref('prefixItems', 'type'),
    "contains": Ref('contains', 'type'),
    "type": Ref('type', 'type_value')
},
    primitives="enum,const,"  # enums and consts
               "required,minProperties,maxProperties,"  # object
               "minItems,maxItems,uniqueItems,minContains,maxContains,"  # array
               "minLength,maxLength,pattern,format,"  # string
               "minimum,exclusiveMinimum,maximum,exclusiveMaximum,multipleOf"  # numbers
)

Ref.refs['additionalProperties'] = DictValidator(primitives='type')

Ref.refs['propertyNames'] = DictValidator(primitives='pattern'),

Ref.refs['properties'] = AnyKeyDictValidator(Ref('property', 'type'))

Ref.refs['type_value'] = PrimitiveValidator()

schema_paths = DictValidator(
    {
        "$vocabulary": AnyKeyDictValidator(primitive),
        "properties": Ref('properties', 'properties')
    }, primitives='$schema,$id,type'
)

model_paths = DictValidator(
    {
        'model': schema_paths
    }
)


def match_schema(stack: ModelBuilderStack):
    sc = model_paths
    parents = [sc]
    for entry in stack.stack:
        key = entry.key
        if key is None:
            continue
        if isinstance(key, int):
            continue
        sc = sc.get(key)
        if not sc.valid:
            return None
        parents.append(sc)
    return parents


def is_schema_element(stack: ModelBuilderStack):
    return match_schema(stack) is not None



from ..builders import OutputBuilder
from ..outputs.toml import TOMLOutput


class InvenioRecordResourcePoetryBuilder(OutputBuilder):
    TYPE = 'invenio_record_resource_poetry'

    def finish(self):
        super().finish()

        output: TOMLOutput = self.builder.get_output(
            'toml',
            'pyproject.toml'
        )

        register_function = self.settings.python.create_blueprint_from_app.rsplit('.', maxsplit=1)

        output.set("tool.poetry.plugins.'invenio_base.api_blueprints'",
                   self.settings.python.record_resource_blueprint_name,
                   f'{register_function[0]}:{register_function[-1]}'
                   )



import sys
import libcst as cst
from jinja2 import Environment, FunctionLoader, pass_context

from oarepo_model_builder.templates import templates
from oarepo_model_builder.outputs import OutputBase
from oarepo_model_builder.utils.cst import MergingTransformer, merge
from oarepo_model_builder.utils.jinja import in_different_package, base_name, package_name, with_defined_prefix
from oarepo_model_builder.utils.verbose import log


class PythonOutput(OutputBase):
    TYPE = 'python'
    cst = None
    original_data = None

    def begin(self):
        try:
            with self.builder.filesystem.open(self.path) as f:
                self.original_data = f.read()

                self.cst = cst.parse_module(self.original_data)
        except FileNotFoundError:
            self.original_data = None
            self.cst = cst.parse_module('')

    @property
    def created(self):
        return self.original_data is None

    def finish(self):
        code = self.cst.code
        if code != self.original_data:
            self.builder.filesystem.mkdir(self.path.parent)
            log(2, 'Saving %s', self.path)
            with self.builder.filesystem.open(self.path, mode='w') as f:
                f.write(code)
            if self.builder.schema.settings.python.use_isort:
                import isort
                config = isort.settings.Config(verbose=False, quiet=True)
                isort.file(self.path, config=config)
            if self.builder.schema.settings.python.use_black:
                import subprocess
                subprocess.call([
                    'black',
                    '-q',
                    '--experimental-string-processing',
                    str(self.path)
                ])

    def merge(self, template_name, context, filters=None):
        # template is a loadable resource
        env = Environment(
            loader=FunctionLoader(lambda tn: templates.get_template(tn, context['settings'])),
            autoescape=False,
        )
        self.register_default_filters(env)
        for filter_name, filter_func in (filters or {}).items():
            env.filters[filter_name] = filter_func

        rendered = env.get_template(template_name).render(context)
        try:
            rendered_cst = cst.parse_module(rendered, config=self.cst.config_for_parsing)
        except:
            print(rendered, file=sys.stderr)
            raise
        self.cst = merge(self.cst, rendered_cst)

    @staticmethod
    def register_default_filters(env):
        env.filters['package_name'] = package_name
        env.filters['base_name'] = pass_context(
            lambda context, value:
            base_name(value)
            if not with_defined_prefix(
                context['settings'].python.always_defined_import_prefixes, value)
            else value
        )
        env.tests['in_different_package'] = pass_context(
            lambda context, value: in_different_package(context['current_package_name'], value)
        )
        env.tests['not_prefixed'] = pass_context(
            lambda context, value: not with_defined_prefix(
                context['settings'].python.always_defined_import_prefixes, value)
        )


class CSTPart:
    def __init__(self, output, path):
        self.output = output
        self.path = path

    @property
    def _cst(self):
        current = self.output.cst
        for p in self.path:
            for c in current.body:
                if isinstance(c, p[0]) and c.name.value == p[1]:
                    current = c
                    break
            else:
                return None
        return current

    @property
    def exists(self):
        return self._cst is not None

    def _create(self, expr):
        self.output.create_cst(
            cst.parse_module(expr, config=self.output.cst.config_for_parsing).body,
            self.path[:-1]
        )


class PythonClass(CSTPart):
    def __init__(self, output, path):
        super().__init__(output, path)

    def create(self, name=None):
        expr = f"""
class {name}:
    pass
        """
        return self._create(expr)



from .invenio_base import InvenioBaseClassPythonBuilder


class InvenioRecordMetadataBuilder(InvenioBaseClassPythonBuilder):
    TYPE = 'invenio_record_metadata'
    class_config = 'record-metadata-class'
    template = 'record-metadata'
